# -*- coding: utf-8 -*-
from .connector import ProxyConnector
from .errors import (
    SocksError, NoAcceptableAuthMethods, UnknownAuthMethod,
    LoginAuthenticationFailed, InvalidServerVersion, InvalidServerReply,
    SocksConnectionError
)
from .helpers import open_connection, create_connection
from .proto import ProxyType

__version__ = '0.1.1'

__all__ = ('ProxyConnector', 'ProxyType',
           'SocksError', 'NoAcceptableAuthMethods', 'UnknownAuthMethod',
           'LoginAuthenticationFailed', 'InvalidServerVersion',
           'InvalidServerReply', 'SocksConnectionError', 'open_connection',
           'create_connection')
